"""Copyright 2020 SCOUT.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# mypy: ignore-errors
from collections.abc import Callable
from datetime import UTC, datetime
from functools import partial
from typing import Any, NamedTuple

# Types
F = Callable[[Any], Any]


class TypeCode(NamedTuple):
    type_obj: F
    type_name: str


def _option_type(datatype: type, data: Any) -> Any:
    if data:
        return datatype(data)
    if datatype in {bool, int, float} and data == 0:
        return datatype(data)
    return None


class CDatetime:
    Year: int
    Month: int
    Day: int
    Hour: int
    Minute: int
    Second: int


def conv_dt(x: CDatetime) -> datetime | None:
    if not x:
        return None
    return datetime(x.Year, x.Month, x.Day, x.Hour, x.Minute, x.Second, tzinfo=UTC)


def conv_obj(x: Any) -> Any:
    return x


adomd_type_map: dict[str, TypeCode] = {
    "System.Boolean": TypeCode(partial(_option_type, bool), bool.__name__),
    "System.DateTime": TypeCode(
        conv_dt,
        datetime.__name__,
    ),
    # "System.Decimal": TypeCode(
    #     lambda x: Decimal.ToDouble(x) if x else None, float.__name__  #
    # ),
    "System.Double": TypeCode(partial(_option_type, float), float.__name__),
    "System.Single": TypeCode(partial(_option_type, float), float.__name__),
    "System.String": TypeCode(partial(_option_type, str), str.__name__),
    "System.Guid": TypeCode(partial(_option_type, str), str.__name__),
    "System.UInt16": TypeCode(partial(_option_type, int), int.__name__),
    "System.UInt32": TypeCode(partial(_option_type, int), int.__name__),
    "System.UInt64": TypeCode(partial(_option_type, int), int.__name__),
    "System.Int16": TypeCode(partial(_option_type, int), int.__name__),
    "System.Int32": TypeCode(partial(_option_type, int), int.__name__),
    "System.Int64": TypeCode(partial(_option_type, int), int.__name__),
    "System.Object": TypeCode(conv_obj, "System.Object"),
}


def convert(datatype: str, data: Any, type_map: dict[str, TypeCode]) -> Any:
    type_to_convert = type_map[datatype]
    return type_to_convert.type_obj(data)



# ruff: noqa: N802, N999, PYI021
from typing import Any

from . import enums

class ConnectionState:
    value__: enums.ConnectionState

class IDataReader:
    @property
    def IsClosed(self) -> bool: ...
    @property
    def FieldCount(self) -> int: ...
    def Close(self) -> None: ...
    def ReadOuterXml(self) -> str: ...
    def GetName(self, idx: int) -> str: ...
    def GetFieldType(self, idx: int) -> ReaderType: ...
    def __getitem__(self, idx: int) -> Any: ...
    def Read(self) -> bool: ...

class XmlReader(IDataReader): ...

class ReaderType:
    """Actually just called Type, but prefix for clarity."""

    def ToString(self) -> str: ...

class AdomdDataReader(IDataReader):
    @property
    def FieldCount(self) -> int: ...

class AdomdCommand:
    def __init__(self, query: str, conn: AdomdConnection) -> None: ...
    def ExecuteXmlReader(self) -> XmlReader: ...
    def ExecuteReader(self) -> AdomdDataReader: ...
    def ExecuteNonQuery(self) -> int: ...

class AdomdConnection:
    @property
    def State(self) -> ConnectionState: ...
    @property
    def ConnectionString(self) -> str: ...
    def __init__(self, conn_str: str) -> None: ...
    def Open(self) -> None: ...
    def Close(self) -> None: ...
    def Dispose(self) -> None: ...
    def Clone(self) -> AdomdConnection: ...

class AdomdErrorResponseException(BaseException):
    Message: str

class AdomdTransaction: ...

class AdomdUnknownResponseException(BaseException):
    Message: str



"""Copyright 2020 SCOUT.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# mypy: ignore-errors
from dataclasses import dataclass

import bs4
import structlog

logger = structlog.get_logger()


@dataclass
class XmlaErrorInfo:
    description: str
    error_code: int
    help_file: str | None
    source: str | None

    def __str__(self) -> str:
        return (
            f"Error Code: {self.error_code}\n"
            f"Description: {self.description}\n"
            f"Source: {self.source}\n"
            f"Help File: {self.help_file}"
        )


class XmlaResponseError(Exception):
    """Exception raised for XMLA error responses."""

    def __init__(
        self,
        errors: list[XmlaErrorInfo],
        full_message: bs4.BeautifulSoup,
    ) -> None:
        message = "XMLA error response"
        error_str = "\n\n".join(str(x) for x in errors)
        message += f"\n\n{error_str}"

        super().__init__(message)
        self.errors = errors
        self.full_message = full_message


def check_errors(xml: bs4.BeautifulSoup) -> None:
    errors = [
        XmlaErrorInfo(
            node["Description"],  # pyright: ignore  # noqa: PGH003
            int(node["ErrorCode"]),  # pyright: ignore  # noqa: PGH003
            node["HelpFile"] or None,  # pyright: ignore  # noqa: PGH003
            node["Source"] or None,  # pyright: ignore  # noqa: PGH003
        )
        for node in xml.find_all("Error")
    ]
    if errors:
        raise XmlaResponseError(
            errors,
            xml,
        )



"""Copyright 2020 SCOUT.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# mypy: ignore-errors
from collections.abc import Iterator
from pathlib import Path
from sys import path
from typing import TYPE_CHECKING, Any, NamedTuple

import clr  # type: ignore[import-untyped]
import structlog

from .c_sharp_type_mapping import adomd_type_map, convert

logger = structlog.get_logger()


path.append(str(Path(__file__).parent)[2:])
clr.AddReference("Microsoft.AnalysisServices.AdomdClient")  # pyright: ignore reportAttributeAccessIssue
from Microsoft.AnalysisServices.AdomdClient import (  # noqa: E402
    AdomdUnknownResponseException,
)

if TYPE_CHECKING:
    from types import TracebackType

    from Microsoft.AnalysisServices.AdomdClient import IDataReader


class Description(NamedTuple):
    name: str
    type_code: str


class Reader:
    _reader: "IDataReader"

    def __init__(self, reader: "IDataReader") -> None:
        self._reader = reader

    def __enter__(self) -> "Reader":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: "TracebackType | None",  # noqa: PYI036
    ) -> None:
        self.close()

    def close(self) -> None:
        self._reader.Close()

    def column_names(self) -> list[str]:
        """Returns the column names of the last executed query."""
        return [self._reader.GetName(i) for i in range(self.field_count)]

    def descriptions(self) -> list[Description]:
        return [
            Description(
                self._reader.GetName(i),
                adomd_type_map[self._reader.GetFieldType(i).ToString()].type_name,
            )
            for i in range(self.field_count)
        ]

    def fetch_many(self, limit: int | None = None) -> list[dict[str, Any]]:
        """Fetches multiple rows from the last executed query.

        Args:
            limit (int | None): The number of rows to fetch. If None, fetches all rows.

        Returns:
            list[dict[str, Any]]: A list of dictionaries representing the rows.

        """
        if limit is not None:
            return [self.fetch_one() for _ in range(limit) if self.read()]
        return list(self.fetch_stream())

    def fetch_one(self) -> dict[str, Any]:
        """Fetches a single row from the last executed query as a dictionary.

        Returns:
            dict[str, Any]: A dictionary representing the row, with column names as keys

        """
        column_names = self.column_names()
        data = self.fetch_one_tuple()
        return dict(zip(column_names, data, strict=False))

    def fetch_one_tuple(self) -> tuple[Any, ...]:
        """Fetches a single row from the last executed query as a tuple.

        Returns:
            tuple[Any, ...]: A tuple representing the row, with C# values converted
                to their appropriate python types.

        """
        return tuple(
            convert(
                self._reader.GetFieldType(i).ToString(),
                self._reader[i],
                adomd_type_map,
            )
            for i in range(self.field_count)
        )

    def fetch_stream(self) -> Iterator[dict[str, Any]]:
        """Fetches the rows from the last executed query as a stream of dictionaries.

        Note:
            You may need to close the reader after fetching the rows if:

            1. You are using a explicit limit that is shorter than the total number of
            rows in the query result
            2. You are tracing the command associated with the reader

            This is because the trace will not create a query end record (since it
            assumes the client is still reading) without explicitly closing the
            reader. The reader can be closed with `self._reader.Close()`

        """
        column_names = self.column_names()
        while self.read():
            yield dict(zip(column_names, self.fetch_one_tuple(), strict=False))

    @property
    def field_count(self) -> int:
        return self._reader.FieldCount

    @property
    def is_closed(self) -> bool:
        return self._reader.IsClosed

    def read(self) -> bool:
        try:
            return self._reader.Read()
        except AdomdUnknownResponseException:
            return False

    def read_outer_xml(self) -> str:
        return self._reader.ReadOuterXml()



# ruff: noqa: N802, PYI021

from collections.abc import Iterator
from enum import IntEnum
from types import TracebackType
from typing import Any, NamedTuple, Self

import bs4

from .Microsoft.AnalysisServices.AdomdClient import AdomdConnection

class ReaderType:
    """Actually just called Type, but prefix for clarity."""

    def ToString(self) -> str: ...

class AdmomdState(IntEnum):
    OPEN = 1
    CLOSED = 0

class IDataReader:
    @property
    def IsClosed(self) -> bool: ...
    def Close(self) -> None: ...
    def ReadOuterXml(self) -> str: ...
    def GetName(self, idx: int) -> str: ...
    def GetFieldType(self, idx: int) -> ReaderType: ...
    def __getitem__(self, idx: int) -> Any: ...
    def Read(self) -> bool: ...

class AdomdErrorResponseException(BaseException):
    Message: str

class Description(NamedTuple):
    name: str
    type_code: str

class Reader:
    _reader: IDataReader

    # No implementation of __init__ since it shouldn't be called directly.
    def __enter__(self) -> Self: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None: ...
    def read(self) -> bool: ...
    def read_outer_xml(self) -> str: ...
    def column_names(self) -> list[str]: ...
    def descriptions(self) -> list[Description]: ...
    def get_row(self) -> tuple[Any, ...]: ...
    @property
    def field_count(self) -> int: ...
    @property
    def is_closed(self) -> bool: ...
    def close(self) -> None: ...
    def fetch_stream(self) -> Iterator[dict[str, Any]]: ...
    def fetch_one(self) -> dict[str, Any]: ...
    def fetch_one_tuple(self) -> tuple[Any, ...]: ...
    def fetch_many(self, limit: int | None = None) -> list[dict[str, Any]]: ...

class Connection:
    conn: AdomdConnection
    def __init__(self, conn_str: str) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None: ...
    def close(self) -> None: ...
    def open(self) -> Self: ...
    @property
    def state(self) -> AdmomdState: ...
    def execute_xml(
        self,
        query: str,
        query_name: str | None = None,
    ) -> bs4.BeautifulSoup: ...
    def execute_dax(self, query: str, query_name: str | None = None) -> Reader: ...

def connect(conn_str: str) -> Connection: ...



"""Copyright 2020 SCOUT.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# mypy: ignore-errors
from pathlib import Path
from sys import path
from typing import TYPE_CHECKING, Self

import bs4
import clr  # type: ignore[import-untyped]
import structlog

from . import utils, xml_error_handling
from .Microsoft.AnalysisServices.enums import ConnectionState
from .reader import Reader

logger = structlog.get_logger()


path.append(str(Path(__file__).parent)[2:])
clr.AddReference("Microsoft.AnalysisServices.AdomdClient")  # pyright: ignore reportAttributeAccessIssue
from Microsoft.AnalysisServices.AdomdClient import (  # noqa: E402
    AdomdCommand,
    AdomdConnection,
    AdomdErrorResponseException,
)

__all__ = [
    "AdomdErrorResponseException",
    "Connection",
]  # needed to keep ruff from cleaning up the exception

if TYPE_CHECKING:
    from types import TracebackType


class Connection:
    conn: AdomdConnection
    """The underlying C# AdomdConnection object."""
    reader: Reader
    """The python reader for the connection, used to execute queries."""

    def __init__(self, conn_str: str) -> None:
        self.conn = AdomdConnection(conn_str)

    def __enter__(self) -> Self:
        if self.state != ConnectionState.Open:
            self.open()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: "TracebackType | None",  # noqa: PYI036
    ) -> None:
        self.close()

    def clone(self) -> "Connection":
        """Clones the connection."""
        return Connection(self.conn.ConnectionString)

    def close(self) -> None:
        """Closes the connection."""
        self.conn.Close()
        self.conn.Dispose()

    def execute_dax(self, query: str, query_name: str | None = None) -> Reader:
        """Executes a DAX query and returns a Reader object.

        Args:
            query (str): The DAX query to execute.
            query_name (str | None): Optional name for the query, used for logging.

        Returns:
            Reader: A Reader object to read the results of the query.

        """
        query_name = query_name or ""
        logger.debug("execute DAX query", query_name=query_name)
        cmd = AdomdCommand(query, self.conn)
        return Reader(cmd.ExecuteReader())

    def execute_non_query(self, query: str, query_name: str | None = None) -> Self:
        """Executes a non-query DAX command.

        Returns:
            Self: The connection object itself for method chaining.

        """
        query_name = query_name or ""
        logger.debug("execute DAX query", query_name=query_name)
        cmd = AdomdCommand(query, self.conn)
        cmd.ExecuteNonQuery()
        return self

    def execute_xml(
        self,
        query: str,
        query_name: str | None = None,
    ) -> bs4.BeautifulSoup:
        query_name = query_name or ""
        logger.debug("execute XML query", query_name=query_name)
        cmd = AdomdCommand(query, self.conn)

        with Reader(cmd.ExecuteXmlReader()) as reader:
            logger.debug("reading query", query_name=query_name)
            lines = [reader.read_outer_xml()]
            while lines[-1] != "":
                lines.append(reader.read_outer_xml())
            ret = bs4.BeautifulSoup("".join(lines), "xml")
            for node in ret.find_all():
                assert isinstance(node, bs4.element.Tag)
                node.name = utils._decode_name(node.name)

        self._check_errors(ret)
        return ret

    def open(self) -> Self:
        """Opens the connection."""
        self.conn.Open()
        return self

    @property
    def state(self) -> ConnectionState:
        """1 = Open, 0 = Closed."""
        return ConnectionState(self.conn.State.value__)

    @staticmethod
    def _check_errors(xml: bs4.BeautifulSoup) -> None:
        xml_error_handling.check_errors(xml)


def connect(conn_str: str) -> Connection:
    """Connects to the given connection string."""
    return Connection(conn_str)



def _is_encoded_char(val: str) -> bool:
    utf_encoded_len = 5
    start, body = val[0], val[1:]
    return (
        len(val) == utf_encoded_len
        and start == "x"
        and all(c in "0123456789ABCDEF" for c in body)
    )


def _decode_name(name: str) -> str:
    """Processes names to decode any encoded characters within an XML-safe string."""
    name_parts = name.split("_")
    for i, e in enumerate(name_parts):
        if _is_encoded_char(e):
            name_parts[i] = chr(int(e[1:], 16))
    return "_".join(name_parts)


__all__ = ["_decode_name"]
